from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _is_negative_zero,
    _is_positive_zero,
    _is_sign_negative_number,
    _is_sign_positive_number,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarFloor(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarFloor":
        return ScalarFloor(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.floor, ScalarFloor
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.floor(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        arg = self._a

        # compute the rounded result that meets the expr bounds
        # rounding uses integer steps, so e.g. floor(...) in [-3.1, 4.7] means
        #  that floor(...) in [-3, 4]
        expr_lower = np.ceil(expr_lower)
        expr_upper = np.floor(expr_upper)

        # compute the arg bound that will round to meet the expr bounds
        # floor rounds down
        # if expr_lower is -0.0, ceil(-0.0) = -0.0, only floor(-0.0) = -0.0
        # if expr_lower is +0.0, ceil(+0.0) = +0.0, there is nothing below +0.0
        #  for which floor(...) = +0.0
        # if expr_upper is -0.0, floor(-0.0) = -0.0, we need to force arg_upper
        #  to -0.0
        # if expr_upper is +0.0, floor(+0.0) = +0.0, and floor(1-eps) = +0.0
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(expr_lower)
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.nextafter(expr_upper + 1, expr_upper)
        )
        arg_upper[_is_negative_zero(expr_upper)] = -0.0

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"floor({self._a!r})"


class ScalarCeil(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarCeil":
        return ScalarCeil(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.ceil, ScalarCeil
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.ceil(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        arg = self._a

        # compute the rounded result that meets the expr bounds
        # rounding uses integer steps, so e.g. ceil(...) in [-3.1, 4.7] means
        #  that ceil(...) in [-3, 4]
        expr_lower = np.ceil(expr_lower)
        expr_upper = np.floor(expr_upper)

        # compute the arg bound that will round to meet the expr bounds
        # ceil rounds up
        # if expr_lower is -0.0, ceil(-0.0) = -0.0, and ceil(-1+eps) = -0.0
        # if expr_lower is +0.0, ceil(+0.0) = +0.0, we need to force arg_lower
        #  to +0.0
        # if expr_upper is -0.0, floor(-0.0) = -0.0, there is nothing above
        #  -0.0 for which ceil(...) = -0.0
        # if expr_upper is +0.0, floor(+0.0) = +0.0, only ceil(+0.0) = +0.0
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.nextafter(expr_lower - 1, expr_lower)
        )
        arg_lower[_is_positive_zero(expr_lower)] = +0.0
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(expr_upper)

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"ceil({self._a!r})"


class ScalarTrunc(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarTrunc":
        return ScalarTrunc(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a, dtype, np.trunc, ScalarTrunc
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.trunc(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        arg = self._a

        # compute the rounded result that meets the expr bounds
        # rounding uses integer steps, so e.g. trunc(...) in [-3.1, 4.7] means
        #  that trunc(...) in [-3, 4]
        expr_lower = np.ceil(expr_lower)
        expr_upper = np.floor(expr_upper)

        # compute the arg bound that will round to meet the expr bounds
        # trunc rounds towards zero
        # if expr_lower is -0.0, ceil(-0.0) = -0.0, and trunc(-1+eps) = -0.0
        # if expr_lower is +0.0, ceil(+0.0) = +0.0, there is nothing below +0.0
        #  for which trunc(...) = +0.0
        # if expr_upper is -0.0, floor(-0.0) = -0.0, there is nothing above
        #  -0.0 for which trunc(...) = -0.0
        # if expr_upper is +0.0, floor(+0.0) = +0.0, and trunc(1-eps) = +0.0
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.nextafter(expr_lower - 1, expr_lower)
        )
        np.copyto(
            arg_lower,
            expr_lower,
            where=_is_sign_positive_number(expr_lower),
            casting="no",
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.nextafter(expr_upper + 1, expr_upper)
        )
        np.copyto(
            arg_upper,
            expr_upper,
            where=_is_sign_negative_number(expr_upper),
            casting="no",
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"trunc({self._a!r})"


class ScalarRoundTiesEven(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr):
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarRoundTiesEven":
        return ScalarRoundTiesEven(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.rint,
            ScalarRoundTiesEven,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.rint(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and round_ties_even(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.rint(argv)

        # compute the rounded result that meets the expr bounds
        # rounding uses integer steps, so e.g. round_ties_even(...) in
        #  [-3.1, 4.7] means that round_ties_even(...) in [-3, 4]
        expr_lower = np.ceil(expr_lower)
        expr_upper = np.floor(expr_upper)

        # estimate the arg bound that will round to meet the expr bounds
        # round_ties_even rounds to the nearest integer, with tie breaks
        #  towards the nearest *even* integer
        # if expr_lower is -0.0, ceil(-0.0) = -0.0, and rint(-0.5) = -0.0
        # if expr_lower is +0.0, ceil(+0.0) = +0.0, we need to force arg_lower
        #  to +0.0
        # if expr_upper is -0.0, floor(-0.0) = -0.0, we need to force arg_upper
        #  to -0.0
        # if expr_upper is +0.0, floor(+0.0) = +0.0, and rint(+0.5) = +0.0
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.subtract(expr_lower, Xs.dtype.type(0.5))
        )
        arg_lower[_is_positive_zero(expr_lower)] = +0.0
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.add(expr_upper, Xs.dtype.type(0.5))
        )
        arg_upper[_is_negative_zero(expr_upper)] = -0.0

        # handle rounding errors in round_ties_even(...) early
        # which can occur since our +-0.5 estimate doesn't account for the even
        #  tie breaking cases
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.rint(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.rint(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"round_ties_even({self._a!r})"



"""
Exception and warning types and helpers to raise exceptions with context.

Exceptions raised in this package carry [`ErrorContext`][.ErrorContext], which
can be accessed through the [`ErrorContextMixin`][.ErrorContextMixin].
"""

__all__ = [
    "ContextLayer",
    "ErrorContext",
    "ctx",
    "ErrorContextMixin",
    "SafeguardTypeContextLayer",
    "ParameterContextLayer",
    "LateBoundParameterContextLayer",
    "IndexContextLayer",
    "TypeCheckError",
    "TypeSetError",
    "LateBoundParameterResolutionError",
    "SafeguardsSafetyBug",
    "QuantityOfInterestRuntimeWarning",
    "Ei",
    "lookup_enum_or_raise",
]

from abc import ABC, abstractmethod
from collections.abc import Set
from contextlib import AbstractContextManager, contextmanager
from enum import Enum
from types import UnionType
from typing import TYPE_CHECKING, Never, Self, TypeVar, final

import numpy as np
from typing_extensions import override  # MSPV 3.12

if TYPE_CHECKING:
    from ..safeguards.abc import Safeguard
    from .bindings import Parameter

Ei = TypeVar("Ei", bound=Enum)
""" Any enum type (invariant). """


class ContextLayer(ABC):
    """
    Context layer abstract base class.
    """

    __slots__: tuple[str, ...] = ()

    @override
    @abstractmethod
    def __str__(self) -> str:
        pass

    @property
    def separator(self) -> str:
        """
        Separator to print before this layer, `'.'` by default.
        """

        return "."


class ErrorContext:
    """
    Context in which an error was raised.

    The context can be added to an exception using the
    `err | ErrorContext(...)` syntax.

    Parameters
    ----------
    *context : ContextLayer
        The layers of the context.
    """

    __slots__: tuple[str, ...] = ("_layers",)
    _layers: tuple[ContextLayer, ...]

    def __init__(self, *context: ContextLayer):
        self._layers = context

    @property
    def layers(self) -> tuple[ContextLayer, ...]:
        """
        The layers of the context, from outermost to innermost.
        """

        return self._layers

    def __ror__(self, other: BaseException) -> BaseException:
        if isinstance(other, ErrorContextMixin):
            other._context = ErrorContext(*self.layers, *other.context.layers)
            return other

        ty = type(other)
        ty_with_context = _EXCEPTIONS_WITH_CONTEXT.get(ty, None)

        if ty_with_context is None:

            def __str__with_context(self) -> str:
                context_str = str(self.context)
                err_str = super(type(self), self).__str__()
                if context_str == "":
                    return err_str
                return f"{context_str}: {err_str}"

            ty_with_context = type(
                ty.__name__,
                (ty, ErrorContextMixin),
                dict(__str__=__str__with_context, __module__=ty.__module__),
            )
            _EXCEPTIONS_WITH_CONTEXT[ty] = ty_with_context

        other_with_context = ty_with_context(*other.args)
        other_with_context._context = self  # type: ignore
        other_with_context.__cause__ = other.__cause__
        other_with_context.__context__ = other.__context__
        other_with_context.__suppress_context__ = other.__suppress_context__
        other_with_context.__traceback__ = other.__traceback__
        if hasattr(other, "__notes__"):
            # only present after add_note() is called
            other_with_context.__notes__ = other.__notes__
        return other_with_context

    @override
    def __str__(self) -> str:
        match self.layers:
            case ():
                return ""
            case (c,):
                return str(c)
            case _:
                c, *cs = self.layers
                acc = [str(c)]
                for c in cs:
                    acc.append(c.separator)
                    acc.append(str(c))
                return "".join(acc)


_EXCEPTIONS_WITH_CONTEXT: dict[type[BaseException], type[BaseException]] = dict()


@final
class _ctxmeta(type):
    @override
    def __ror__(self, other: BaseException) -> BaseException:  # type: ignore
        return other | ErrorContext()


@final
class ctx(metaclass=_ctxmeta):
    """
    Singleton error context type with which additional context scopes can be entered.

    An exception can be prepared for receiving context using the `err | ctx`
    syntax.
    """

    __slots__: tuple[str, ...] = ()

    def __new__(cls) -> Self:
        raise TypeError(f"{cls} is a singleton") | ctx

    @contextmanager
    @staticmethod
    def layer(layer: ContextLayer):
        """
        Context manager that adds one `layer` of context to any exception that
        is raised within.

        Parameters
        ----------
        layer : ContextLayer
            The layer of context to add.
        """

        try:
            yield
        except Exception as err:
            err2 = err | ErrorContext(layer)
            if isinstance(err, ErrorContextMixin):
                raise
            raise err2

    @staticmethod
    def safeguard(safeguard: "Safeguard") -> AbstractContextManager[None]:
        """
        Context manager that adds one layer of context, the type of a
        `safeguard`, to any exception that is raised within.

        The added context layer will be a
        [`SafeguardTypeContextLayer`][...SafeguardTypeContextLayer].

        Parameters
        ----------
        safeguard : Safeguard
            A safeguard whose type to add as a layer of context.
        """

        return ctx.layer(SafeguardTypeContextLayer(type(safeguard)))

    @staticmethod
    def safeguardty(safeguard: type["Safeguard"]) -> AbstractContextManager[None]:
        """
        Context manager that adds one layer of context, the `safeguard` type,
        to any exception that is raised within.

        The added context layer will be a
        [`SafeguardTypeContextLayer`][...SafeguardTypeContextLayer].

        Parameters
        ----------
        safeguard : type[Safeguard]
            The type of a safeguard to add as a layer of context.
        """

        return ctx.layer(SafeguardTypeContextLayer(safeguard))

    @staticmethod
    def parameter(name: str) -> AbstractContextManager[None]:
        """
        Context manager that adds one layer of context, the `name` of a
        parameter, to any exception that is raised within.

        The added context layer will be a
        [`ParameterContextLayer`][...ParameterContextLayer].

        Parameters
        ----------
        name : str
            The name of a parameter to add as a layer of context.
        """

        return ctx.layer(ParameterContextLayer(name))

    @staticmethod
    def late_bound_parameter(name: "Parameter") -> AbstractContextManager[None]:
        """
        Context manager that adds one layer of context, the `name` of a
        late-bound parameter, to any exception that is raised within.

        The added context layer will be a
        [`LateBoundParameterContextLayer`][...LateBoundParameterContextLayer].

        Parameters
        ----------
        name : Parameter
            The name of a late-bound parameter to add as a layer of context.
        """

        return ctx.layer(LateBoundParameterContextLayer(name))

    @staticmethod
    def index(index: int) -> AbstractContextManager[None]:
        """
        Context manager that adds one layer of context, an `index`, to any
        exception that is raised within.

        The added context layer will be an
        [`IndexContextLayer`][...IndexContextLayer].

        Parameters
        ----------
        index : int
            An index to add as a layer of context.
        """

        return ctx.layer(IndexContextLayer(index))

    @staticmethod
    def __ror__(other: BaseException) -> BaseException:  # type: ignore
        return other | ctx


class ErrorContextMixin:
    """
    Mixin for exceptions that have [`ErrorContext`][compression_safeguards.utils.error.ErrorContext].

    The `context` can be accessed after checking
    `isinstance(err, ErrorContextMixin)`.
    """

    # cannot use slots since they are incompatible with multiple inheritance
    # __slots__: tuple[str, ...] = ("_context",)

    _context: ErrorContext

    @property
    def context(self) -> ErrorContext:
        """
        The context in which this exception was raised.
        """

        try:
            return self._context
        except AttributeError:
            context = ErrorContext()
            self._context = context
            return context


class SafeguardTypeContextLayer(ContextLayer):
    """
    Safeguard type context layer.

    Parameters
    ----------
    safeguard : type[Safeguard]
        The safeguard type.
    """

    __slots__: tuple[str, ...] = ("_safeguard",)
    __match_args__: tuple[str, ...] = ("safeguard",)
    _safeguard: type["Safeguard"]

    def __init__(self, safeguard: type["Safeguard"]):
        self._safeguard = safeguard

    @property
    def safeguard(self) -> type["Safeguard"]:
        """
        The safeguard type.
        """

        return self._safeguard

    @override
    def __str__(self) -> str:
        return self._safeguard.kind

    @override
    def __eq__(self, value: object, /) -> bool:
        return (
            isinstance(value, SafeguardTypeContextLayer)
            and value.safeguard == self.safeguard
        )


class ParameterContextLayer(ContextLayer):
    """
    Parameter context layer.

    Parameters
    ----------
    parameter : str
        The name of the parameter.
    """

    __slots__: tuple[str, ...] = ("_parameter",)
    __match_args__: tuple[str, ...] = ("parameter",)
    _parameter: str

    def __init__(self, parameter: str):
        self._parameter = parameter

    @property
    def parameter(self) -> str:
        """
        The name of the parameter.
        """

        return self._parameter

    @override
    def __str__(self) -> str:
        return self._parameter

    @override
    def __eq__(self, value: object, /) -> bool:
        return (
            isinstance(value, ParameterContextLayer)
            and value.parameter == self.parameter
        )


class LateBoundParameterContextLayer(ContextLayer):
    """
    Late-bound parameter context layer.

    Parameters
    ----------
    parameter : Parameter
        The late-bound parameter.
    """

    __slots__: tuple[str, ...] = ("_parameter",)
    __match_args__: tuple[str, ...] = ("parameter",)
    _parameter: "Parameter"

    def __init__(self, parameter: "Parameter"):
        self._parameter = parameter

    @property
    def parameter(self) -> "Parameter":
        """
        The late-bound parameter.
        """

        return self._parameter

    @override
    def __str__(self) -> str:
        return str(self._parameter)

    @property
    @override
    def separator(self) -> str:
        """
        `'='` separator to print before this layer.
        """

        return "="

    @override
    def __eq__(self, value: object, /) -> bool:
        return (
            isinstance(value, LateBoundParameterContextLayer)
            and value.parameter == self.parameter
        )


class IndexContextLayer(ContextLayer):
    """
    Index context layer.

    Parameters
    ----------
    index : int
        The index.
    """

    __slots__: tuple[str, ...] = ("_index",)
    __match_args__: tuple[str, ...] = ("index",)
    _index: int

    def __init__(self, index: int):
        self._index = index

    @property
    def index(self) -> int:
        """
        The index.
        """

        return self._index

    @override
    def __str__(self) -> str:
        return f"[{self.index}]"

    @property
    @override
    def separator(self) -> str:
        """
        Empty `""` separator to print before this layer.
        """

        return ""

    @override
    def __eq__(self, value: object, /) -> bool:
        return isinstance(value, IndexContextLayer) and value.index == self.index


class TypeCheckError(TypeError):
    """
    [`TypeError`][TypeError] that is raised when a value fails a type check.

    Parameters
    ----------
    expected : type | UnionType
        The expected type or type union.
    found : object
        The value that failed the type check.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(
        self,
        expected: type | UnionType,
        found: object,
    ) -> None:
        super().__init__(expected, found)

    # TODO: once a TypeAssert exists for Python, return it
    @classmethod
    def check_instance_or_raise(
        cls, obj: object, expected: type | UnionType
    ) -> None | Never:
        """
        Check `isinstance(obj, expected)` or raise a type check error, with
        context.

        Parameters
        ----------
        obj : object
            The value to type check.
        expected : type | UnionType
            The expected type or type union.

        Raises
        ------
        TypeCheckError
            if `obj` is not an instance of `expected`.
        """

        if isinstance(obj, expected):
            return None
        raise cls(expected, obj) | ctx

    @property
    def expected(self) -> type | UnionType:
        """
        The expected type or type union.
        """

        (expected, _found) = self.args
        return expected

    @property
    def found(self) -> object:
        """
        The value that failed the type check.
        """

        (_expected, found) = self.args
        return found

    @override
    def __str__(self) -> str:
        return f"expected {self.expected} but found {self.found} of type {type(self.found)}"


class TypeSetError(TypeError):
    """
    [`TypeError`][TypeError] that is raised when a type is not in a type set.

    Parameters
    ----------
    expected : type | UnionType
        The expected type or type union.
    found : type
        The type that failed the type set check.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(self, expected: type | UnionType, found: type):
        super().__init__(expected, found)

    # TODO: once a TypeAssert exists for Python, return it
    @classmethod
    def check_or_raise(cls, ty: type, expected: type | UnionType) -> None | Never:
        """
        Check `issubclass(ty, expected)` or raise a type set error, with
        context.

        Parameters
        ----------
        ty : type
            The type to check.
        expected : type | UnionType
            The expected type or type union.

        Raises
        ------
        TypeSetError
            if `ty` is not a subclass of `expected`.
        """

        if issubclass(ty, expected):
            return None
        raise cls(expected, ty) | ctx

    @classmethod
    def check_dtype_or_raise(
        cls, dtype: np.dtype, supported: Set[np.dtype]
    ) -> None | Never:
        """
        Check if `dtype` is in the set of `supported` data types or raise a
        type set error, with context.

        Parameters
        ----------
        dtype : np.dtype
            The data type to check.
        supported : Set[np.dtype]
            The set of supported data types.

        Raises
        ------
        TypeSetError
            if `dtype` is not in `supported`.
        """

        if dtype in supported:
            return None

        expected: type | UnionType = Never  # type: ignore
        for s in sorted(supported, key=lambda d: d.name):
            expected |= s.type

        raise cls(expected, dtype.type) | ctx

    @property
    def expected(self) -> type | UnionType:
        """
        The expected type or type union.
        """

        (expected, _found) = self.args
        return expected

    @property
    def found(self) -> type:
        """
        The type that failed the type set check.
        """

        (_expected, found) = self.args
        return found

    @override
    def __str__(self) -> str:
        return f"expected {self.expected} but found {self.found}"


class LateBoundParameterResolutionError(KeyError):
    """
    [`KeyError`][KeyError] that is raised when late-bound parameter resolution fails because one or more parameters are missing or extraneous.

    Parameters
    ----------
    expected : Set[Parameter]
        The set of expected late-bound parameters.
    provided : Set[Parameter]
        The type that failed the type set check.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(self, expected: Set["Parameter"], provided: Set["Parameter"]):
        assert expected != provided
        super().__init__(expected, provided)

    @staticmethod
    def check_or_raise(
        expected: Set["Parameter"], provided: Set["Parameter"]
    ) -> None | Never:
        """
        Check if the `expected` set of late-bound parameters matches the
        `provided` set or raise a late-bound parameter resolution error, with
        context.

        Parameters
        ----------
        expected : Set[Parameter]
            The expected set of late-bound parameters.
        provided : Set[Parameter]
            The provided set of late-bound parameters.

        Raises
        ------
        LateBoundParameterResolutionError
            if `provided` is not equal to `expected`.
        """

        if expected == provided:
            return None
        raise LateBoundParameterResolutionError(expected, provided) | ctx

    @property
    def expected(self) -> Set["Parameter"]:
        """
        The expected set of late-bound parameters.
        """

        (expected, _provided) = self.args
        return expected

    @property
    def provided(self) -> Set["Parameter"]:
        """
        The provided set of late-bound parameters.
        """

        (_expected, provided) = self.args
        return provided

    @property
    def missing(self) -> Set["Parameter"]:
        """
        The missing (expected but not provided) set of late-bound parameters.
        """

        return self.expected - self.provided

    @property
    def extraneous(self) -> Set["Parameter"]:
        """
        The extraneous (provided but not expected) set of late-bound parameters.
        """

        return self.provided - self.expected

    @override
    def __str__(self) -> str:
        missing = self.missing
        missing_str = (
            "missing late-bound parameter"
            + ("s " if len(missing) > 1 else " ")
            + ", ".join(f"`{p}`" for p in sorted(missing))
        )

        extraneous = self.extraneous
        extraneous_str = (
            "extraneous late-bound parameter"
            + ("s " if len(extraneous) > 1 else " ")
            + ", ".join(f"`{p}`" for p in sorted(extraneous))
        )

        if len(missing) <= 0:
            return extraneous_str

        if len(extraneous) <= 0:
            return missing_str

        return f"{missing} and {extraneous}"


class SafeguardsSafetyBug(RuntimeError):
    """
    [`RuntimeError`][RuntimeError] that is raised when a fatal safety bug occurs.

    A fatal safety bug occurs when the safeguards are unable to provide the
    requested safety requirement.

    By raising this error, the `compression-safeguards` avoid violating the
    safety requirements.

    When this error is raised, it is a fatal bug in the implementation of the
    `compression-safeguards`, which should be reported at
    <https://github.com/juntyr/compression-safeguards/issues>.

    Parameters
    ----------
    message : str
        The error message.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.add_note(
            "This is a fatal bug in the implementation of the "
            + "`compression-safeguards`. Please report it at "
            + "<https://github.com/juntyr/compression-safeguards/issues>."
        )


class QuantityOfInterestRuntimeWarning(RuntimeWarning):
    """
    [`RuntimeWarning`][RuntimeWarning] that is raised when a recoverable quantity of interest sanity check fails.

    A quantity of interest safeguard raises this warning if it fails at
    providing the requested safety requirement using regular means but is still
    able to recover and uphold the requirement in the end.

    When this warning is raised, it is a non-fatal bug in the implementation of
    the `compression-safeguards`, which should be reported at
    <https://github.com/juntyr/compression-safeguards/issues>.

    Parameters
    ----------
    message : str
        The warning message.
    """

    __slots__: tuple[str, ...] = ()

    def __init__(self, message: str) -> None:
        super().__init__(message)

        self.add_note(
            "This is a non-fatal bug in the implementation of the "
            + "`compression-safeguards`. Please report it at "
            + "<https://github.com/juntyr/compression-safeguards/issues>."
        )


def lookup_enum_or_raise(
    enum: type[Ei], name: str, error: type[Exception] = ValueError
) -> Ei | Never:
    """
    Look up and return the `enum` member with the given `name` or raise an `error`, with context.

    Parameters
    ----------
    enum : type[Ei]
        The enum in which the `name` is looked up.
    name : str
        The enum member name to look up.
    error : type[Exception]
        The type of error that is raised if the `name` is not a member of the
        `enum`.

    Returns
    -------
    member : Ei
        The `enum` member with the given `name`.

    Raises
    ------
    error
        if the `name` is not a member of the `enum`.
    """

    if name in enum.__members__:
        return enum.__members__[name]

    raise (
        error(
            f"unknown {enum.__name__} {name!r}, use one of "
            + f"{', '.join(repr(m) for m in enum.__members__)}"
        )
        | ctx
    )



from collections.abc import Callable, Mapping
from typing import overload

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import _ensure_array
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarIsFinite(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarIsFinite":
        return ScalarIsFinite(a)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            lambda x: classify_to_dtype(np.isfinite, x, dtype),
            ScalarIsFinite,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return classify_to_dtype(np.isfinite, self._a.eval(Xs, late_bound), Xs.dtype)

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg
        arg = self._a
        argv = arg.eval(Xs, late_bound)

        fmax = np.finfo(Xs.dtype).max

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, isfinite(arg) = True, so arg must be finite
        # if expr_upper < 1, isfinite(arg) = False, arg must stay non-finite
        # otherwise, isfinite(arg) in [True, False] and arg can be anything
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(Xs.shape[:1], -fmax)
        arg_lower[np.less_equal(expr_lower, 0)] = -np.inf
        np.copyto(arg_lower, argv, where=np.less(expr_upper, 1), casting="no")

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(Xs.shape[:1], fmax)
        arg_upper[np.less_equal(expr_lower, 0)] = np.inf
        np.copyto(arg_upper, argv, where=np.less(expr_upper, 1), casting="no")

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"isfinite({self._a!r})"


class ScalarIsInf(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarIsInf":
        return ScalarIsInf(a)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            lambda x: classify_to_dtype(np.isinf, x, dtype),
            ScalarIsInf,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return classify_to_dtype(np.isinf, self._a.eval(Xs, late_bound), Xs.dtype)

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg
        arg = self._a
        argv = arg.eval(Xs, late_bound)

        fmax = np.finfo(Xs.dtype).max

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, isinf(arg) = True, so arg must stay infinite
        # if expr_upper < 0, isinf(arg) = False, arg must be non-infinite
        # otherwise, isinf(arg) in [True, False] and arg can be anything
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(Xs.shape[:1], -fmax)
        arg_lower[np.greater_equal(expr_upper, 1)] = -np.inf
        np.copyto(arg_lower, argv, where=np.greater(expr_lower, 0), casting="no")

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(Xs.shape[:1], fmax)
        arg_upper[np.greater_equal(expr_upper, 1)] = np.inf
        np.copyto(arg_upper, argv, where=np.greater(expr_lower, 0), casting="no")

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"isinf({self._a!r})"


class ScalarIsNaN(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarIsNaN":
        return ScalarIsNaN(a)

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            lambda x: classify_to_dtype(np.isnan, x, dtype),
            ScalarIsNaN,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return classify_to_dtype(np.isnan, self._a.eval(Xs, late_bound), Xs.dtype)

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg
        arg = self._a
        argv = arg.eval(Xs, late_bound)

        # by the precondition, expr_lower <= self.eval(Xs) <= expr_upper
        # if expr_lower > 0, isnan(arg) = True, so arg must stay NaN
        # if expr_upper < 0, isnan(arg) = False, arg must be non-NaN
        # otherwise, isnan(arg) in [True, False] and arg can be anything
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], Xs.dtype.type(-np.inf)
        )
        np.copyto(arg_lower, argv, where=np.greater(expr_lower, 0), casting="no")

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
            Xs.shape[:1], Xs.dtype.type(np.inf)
        )
        np.copyto(arg_upper, argv, where=np.greater(expr_lower, 0), casting="no")

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"isnan({self._a!r})"


@overload
def classify_to_dtype(
    classify: Callable[
        [np.ndarray[tuple[Ps], np.dtype[F]]], np.ndarray[tuple[Ps], np.dtype[np.bool]]
    ],
    a: np.ndarray[tuple[Ps], np.dtype[F]],
    dtype: np.dtype[F],
) -> np.ndarray[tuple[Ps], np.dtype[F]]: ...


@overload
def classify_to_dtype(
    classify: Callable[[Fi], bool],
    a: Fi,
    dtype: np.dtype[Fi],
) -> Fi: ...


def classify_to_dtype(classify, a, dtype):
    c = classify(a)

    if not isinstance(c, np.ndarray):
        return _ensure_array(c).astype(dtype, casting="safe")[()]

    return c.astype(dtype, casting="safe")



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _broadcast_to,
    _ensure_array,
    _is_negative_zero,
    _is_positive_zero,
    _is_sign_negative_number,
    _is_sign_positive_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _stack,
    _where,
)
from ....utils.bindings import Parameter
from ..bound import (
    checked_data_bounds,
    guarantee_arg_within_expr_bounds,
    guarantee_stacked_arg_within_expr_bounds,
)
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant
from .literal import Number


class ScalarPower(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    def __new__(cls, a: AnyExpr, b: AnyExpr) -> "ScalarPower | Number":  # type: ignore[misc]
        if isinstance(a, Number) and isinstance(b, Number):
            # symbolical constant propagation for int ** int
            # where the exponent is non-negative and the result thus is an int
            ai, bi = a.as_int(), b.as_int()
            if (ai is not None) and (bi is not None):
                if bi >= 0:
                    return Number.from_symbolic_int(ai**bi)
        return super().__new__(cls)

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarPower | Number":
        return ScalarPower(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a, self._b, dtype, np.power, ScalarPower
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.power(self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound))

    @override
    @checked_data_bounds
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        a_const = not self._a.has_data
        b_const = not self._b.has_data
        assert not (a_const and b_const), "constant power has no data bounds"

        # evaluate a, b, and power(a, b)
        a, b = self._a, self._b
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)
        exprv: np.ndarray[tuple[Ps], np.dtype[F]] = np.power(av, bv)

        # powers of negative numbers are just too tricky, so we just force
        #  av and bv to remain the same if av < 0 and handle av = 0 by cases
        # powers of sign-positive numbers cannot produce sign-negative numbers,
        #  so clamp expr_lower to >= +0.0 if av >= +0.0
        # TODO: when optimising for a ** int, this clamping should be disabled
        expr_lower = _ensure_array(expr_lower, copy=True)
        expr_lower[_is_sign_positive_number(av) & (expr_lower <= 0)] = +0.0

        b_lower: np.ndarray[tuple[Ps], np.dtype[F]]
        b_upper: np.ndarray[tuple[Ps], np.dtype[F]]

        if a_const:
            av_log = np.log(av)

            # apply the inverse function to get the bounds on b
            # if b_lower == bv and bv == -0.0, we need to guarantee that
            #  b_lower is also -0.0, same for b_upper
            b_lower = _ensure_array(
                _minimum_zero_sign_sensitive(bv, np.divide(np.log(expr_lower), av_log))
            )
            b_upper = _ensure_array(
                _maximum_zero_sign_sensitive(bv, np.divide(np.log(expr_upper), av_log))
            )

            # we need to force bv if expr_lower == expr_upper
            np.copyto(b_lower, bv, where=(expr_lower == expr_upper), casting="no")
            np.copyto(b_upper, bv, where=(expr_lower == expr_upper), casting="no")

            smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

            # handle 0 ** bv
            # - +-0 ** +-0 = 1 -> discontinuous, force bv = +-0
            # - +0 ** (>0) = +0 -> allow any bv > 0
            # - +0 ** (<0) = -inf -> allow any bv < 0
            # - -0 ** (>0) = +-0 -> too tricky, just keep both av and bv the same
            # - -0 ** (<0) = +-inf -> too tricky, just keep both av and bv the same
            np.copyto(b_lower, bv, where=_is_negative_zero(av), casting="no")
            np.copyto(b_upper, bv, where=_is_negative_zero(av), casting="no")
            b_lower[(av == 0) & (bv == 0)] = -0.0
            b_upper[(av == 0) & (bv == 0)] = +0.0
            b_lower[_is_positive_zero(av) & (bv > 0)] = smallest_subnormal
            b_upper[_is_positive_zero(av) & (bv > 0)] = np.inf
            b_lower[_is_positive_zero(av) & (bv < 0)] = -np.inf
            b_upper[_is_positive_zero(av) & (bv < 0)] = -smallest_subnormal

            # handle +inf ** bv
            # - +inf ** +-0 = 1 -> discontinuous, force bv = +-0
            # - +inf ** (>0) = +inf -> allow any bv > 0
            # - +inf ** (<0) = +0 -> allow any bv < 0
            b_lower[np.isinf(av) & (bv == 0)] = -0.0
            b_upper[np.isinf(av) & (bv == 0)] = +0.0
            b_lower[np.isinf(av) & (bv > 0)] = smallest_subnormal
            b_upper[np.isinf(av) & (bv > 0)] = np.inf
            b_lower[np.isinf(av) & (bv < 0)] = -np.inf
            b_upper[np.isinf(av) & (bv < 0)] = -smallest_subnormal

            # handle NaN ** bv
            # - NaN ** +-0 = 1 -> discontinuous, force bv = +-0
            # - NaN ** (<0>) = NaN -> allow any bv != 0
            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future
            b_lower[np.isnan(av) & (bv == 0)] = -0.0
            b_upper[np.isnan(av) & (bv == 0)] = +0.0
            b_lower[np.isnan(av) & (bv > 0)] = smallest_subnormal
            b_upper[np.isnan(av) & (bv > 0)] = np.inf
            b_lower[np.isnan(av) & (bv < 0)] = -np.inf
            b_upper[np.isnan(av) & (bv < 0)] = -smallest_subnormal

            # 1 ** [-NaN, -inf, +inf, +NaN] = 1 -> allow any bv
            b_lower[(av == 1)] = -np.inf
            b_upper[(av == 1)] = np.inf

            # powers of negative numbers are just too tricky, so force bv
            np.copyto(b_lower, bv, where=(av < 0), casting="no")
            np.copyto(b_upper, bv, where=(av < 0), casting="no")

            # handle rounding errors in power(a, log(..., base=a)) early
            b_lower = guarantee_arg_within_expr_bounds(
                lambda b_lower: np.power(av, b_lower),
                exprv,
                bv,
                b_lower,
                expr_lower,
                expr_upper,
            )
            b_upper = guarantee_arg_within_expr_bounds(
                lambda b_upper: np.power(av, b_upper),
                exprv,
                bv,
                b_upper,
                expr_lower,
                expr_upper,
            )

            return b.compute_data_bounds(
                b_lower,
                b_upper,
                Xs,
                late_bound,
            )

        a_lower: np.ndarray[tuple[Ps], np.dtype[F]]
        a_upper: np.ndarray[tuple[Ps], np.dtype[F]]

        if b_const:
            # TODO: optimise bounds for a ** int

            # apply the inverse function to get the bounds on a
            # if a_lower == av and av == -0.0, we need to guarantee that
            #  a_lower is also -0.0, same for a_upper
            a_lower = _ensure_array(
                _minimum_zero_sign_sensitive(
                    av, np.power(expr_lower, np.reciprocal(bv))
                )
            )
            a_upper = _maximum_zero_sign_sensitive(
                av, np.power(expr_upper, np.reciprocal(bv))
            )

            # we need to force av if expr_lower == expr_upper
            np.copyto(a_lower, av, where=(expr_lower == expr_upper), casting="no")
            np.copyto(a_upper, av, where=(expr_lower == expr_upper), casting="no")

            # handle +-0 ** bv -> keep av the same
            #  special-cases are optimised below
            np.copyto(a_lower, av, where=(av == 0), casting="no")
            np.copyto(a_upper, av, where=(av == 0), casting="no")

            # [-NaN, -inf, +inf, +NaN] ** +-0 = 1 -> allow any av
            a_lower[(bv == 0)] = -np.inf
            a_upper[(bv == 0)] = np.inf

            one_plus_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(2))
            one_minus_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(0))

            # handle av ** +-inf
            # - 1 ** +-inf = 1 -> discontinuous, force av = 1
            # - (+-0<1) ** +inf = +0 -> allow any -0.0 <= av < 1
            # - (>1) ** +inf = +inf -> allow any av > 1
            # - (+-0<1) ** -inf = +inf -> allow any -0.0 <= av < 1
            # - (>1) ** -inf = +0 -> allow any av > 1
            # 1 ** +-inf = 1, so force av = 1
            a_lower[(av == 1) & np.isinf(bv)] = 1
            a_upper[(av == 1) & np.isinf(bv)] = 1
            a_lower[(av > 1) & np.isinf(bv)] = one_plus_eps
            a_upper[(av > 1) & np.isinf(bv)] = np.inf
            a_lower[(av < 1) & np.isinf(bv)] = -0.0
            a_upper[(av < 1) & np.isinf(bv)] = one_minus_eps

            # handle av ** NaN
            # - 1 ** NaN = 1 -> discontinuous, force av = 1
            # - (<1>) ** NaN = NaN -> allow any av != 1
            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future
            a_lower[(av == 1) & np.isnan(bv)] = 1
            a_upper[(av == 1) & np.isnan(bv)] = 1
            a_lower[(av > 1) & np.isnan(bv)] = one_plus_eps
            a_upper[(av > 1) & np.isnan(bv)] = np.inf
            a_lower[(av < 1) & np.isnan(bv)] = -np.inf
            a_upper[(av < 1) & np.isnan(bv)] = one_minus_eps

            # powers of negative numbers are just too tricky, so force av
            np.copyto(a_lower, av, where=(av < 0), casting="no")
            np.copyto(a_upper, av, where=(av < 0), casting="no")

            # handle rounding errors in power(power(..., 1/b), b) early
            a_lower = guarantee_arg_within_expr_bounds(
                lambda a_lower: np.power(a_lower, bv),
                exprv,
                av,
                a_lower,
                expr_lower,
                expr_upper,
            )
            a_upper = guarantee_arg_within_expr_bounds(
                lambda a_upper: np.power(a_upper, bv),
                exprv,
                av,
                a_upper,
                expr_lower,
                expr_upper,
            )

            return a.compute_data_bounds(
                a_lower,
                a_upper,
                Xs,
                late_bound,
            )

        one_plus_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(2))
        one_minus_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(0))

        expr_upper = _ensure_array(expr_upper, copy=True)

        # if neither a nor b is const, we need to split the data bound, and
        #  we use that, symbolically, a ** b = exp( b * ln(a) ) for a > 0
        # for av = +-0.0, we keep av the same and handle the cases for bv
        # powers of negative numbers are just too tricky, in general, e.g. can
        #  change sign or suddenly become NaN with negative exponents, so we
        #  just force av and bv to remain the same if av < 0
        # since we know how to derive data bounds for exp and ln and how to
        #  split the data bound for a multiplication, we can combine their
        #  derivations to split the data bound on a ** b into bounds on a and b
        # after the split, we check if our rewrite-based derivation works in
        #  practice by checking against the original power(a, b) function
        # in the data bound split for multiplication, the sign of the two terms
        #  ln(a) and b is kept the same to keep the sign of the product the
        #  same, meaning the sign of b and a <1 vs =1 vs >1 have to remain the
        #  same as well
        # exprv <1 vs =1 vs >1 is thus enforced to remain the same, and we can
        #  adjust the expr bounds to reflect that
        expr_lower[exprv == 1] = 1
        expr_upper[exprv == 1] = 1
        expr_lower[_is_sign_positive_number(av) & (exprv > 1) & (expr_lower <= 1)] = (
            one_plus_eps
        )
        expr_upper[_is_sign_positive_number(av) & (exprv < 1) & (expr_upper >= 1)] = (
            one_minus_eps
        )

        exprv_log = np.log(exprv)
        expr_log_lower = np.log(expr_lower)
        expr_log_upper = np.log(expr_upper)

        # we simplify the data bounds split for ln(a) * b by not allowing the
        #  sign of ln(a) * b to change
        # we further simplify the code by working with
        #  abs(ln(a) * b) = abs(ln(a)) * abs(b)
        # and only taking the sign of ln(a) and b into account after the split
        expr_log_abs_lower, expr_log_abs_upper = (
            _where(
                _is_sign_negative_number(exprv_log), -expr_log_upper, expr_log_lower
            ),
            _where(
                _is_sign_negative_number(exprv_log), -expr_log_lower, expr_log_upper
            ),
        )

        av_log = np.log(av)
        av_log_abs = np.abs(av_log)
        bv_abs = np.abs(bv)

        # we use |ln(a)| * |b| here instead of |ln(a ** b)| since the former
        #  behaves better when a**b is rounded to zero but a != 0
        # in particular, the latter would start with ln(0) = -inf, while the
        #  former can work with finite logarithm results
        exprv_log_abs = np.multiply(av_log_abs, bv_abs)

        fmax = np.finfo(Xs.dtype).max
        smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

        # we are given l <= e <= u, which we translate into e/lf <= e <= e*uf
        # - if the factor is infinite, we limit it to fmax
        # - if the factor is NaN, e.g. from 0/0, we set the factor to 1
        # finally we split the factor geometrically in two using the sqrt
        expr_log_abs_lower_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(exprv_log_abs, expr_log_abs_lower)
        )
        expr_log_abs_lower_factor[np.isinf(expr_log_abs_lower_factor)] = fmax
        np.sqrt(expr_log_abs_lower_factor, out=expr_log_abs_lower_factor)
        expr_log_abs_lower_factor[np.isnan(expr_log_abs_lower_factor)] = 1

        expr_log_abs_upper_factor: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.divide(
                expr_log_abs_upper,
                # we avoid division by zero here
                # - if exprv_log_abs = 0 and expr_log_abs_upper = 0, factor = 0 works
                # - if exprv_log_abs = 0 and expr_log_abs_upper > 0, factor is large
                # - if exprv_log_abs > 0 and expr_log_abs_upper > 0, works as normal
                _maximum_zero_sign_sensitive(exprv_log_abs, smallest_subnormal),
            )
        )
        expr_log_abs_upper_factor[np.isinf(expr_log_abs_upper_factor)] = fmax
        np.sqrt(expr_log_abs_upper_factor, out=expr_log_abs_upper_factor)
        expr_log_abs_upper_factor[np.isnan(expr_log_abs_upper_factor)] = 1

        # TODO: for infinite ln(a) * b that includes finite bounds, allow them

        # compute the bounds for abs(ln(a)) and abs(b)
        a_log_abs_lower = np.divide(av_log_abs, expr_log_abs_lower_factor)
        a_log_abs_upper = np.multiply(av_log_abs, expr_log_abs_upper_factor)

        b_abs_lower = np.divide(bv_abs, expr_log_abs_lower_factor)
        b_abs_upper = np.multiply(bv_abs, expr_log_abs_upper_factor)

        # derive the bounds on ln(a) and b based on the bounds on abs(ln(a))
        #  and abs(b), and on a based on ln(a)
        a_log_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _where(
            _is_sign_negative_number(av_log), -a_log_abs_upper, a_log_abs_lower
        )
        a_log_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _where(
            _is_sign_negative_number(av_log), -a_log_abs_lower, a_log_abs_upper
        )

        # if a_lower == av and av == -0.0, we need to guarantee
        #  that a_lower is also -0.0, same for a_upper
        a_lower = _ensure_array(_minimum_zero_sign_sensitive(av, np.exp(a_log_lower)))
        a_upper = _ensure_array(_maximum_zero_sign_sensitive(av, np.exp(a_log_upper)))

        # if b_lower == bv and bv == -0.0, we need to guarantee
        #  that b_lower is also -0.0, same for b_upper
        b_lower = _where(_is_sign_negative_number(bv), -b_abs_upper, b_abs_lower)
        b_lower = _ensure_array(_minimum_zero_sign_sensitive(bv, b_lower))
        b_upper = _where(_is_sign_negative_number(bv), -b_abs_lower, b_abs_upper)
        b_upper = _ensure_array(_maximum_zero_sign_sensitive(bv, b_upper))

        # we need to force av and bv if expr_lower == expr_upper
        np.copyto(a_lower, av, where=(expr_lower == expr_upper), casting="no")
        np.copyto(a_upper, av, where=(expr_lower == expr_upper), casting="no")
        np.copyto(b_lower, bv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(b_upper, bv, where=(expr_lower == expr_upper), casting="no")

        # 1 ** [-NaN, -inf, +inf, +NaN] = 1 -> force av = 1 and allow any bv
        a_lower[av == 1] = 1
        a_upper[av == 1] = 1
        b_lower[av == 1] = -np.inf
        b_upper[av == 1] = +np.inf

        # (<1>) ** NaN = NaN -> allow any av != 1
        # TODO: an interval union could represent that the two disjoint
        #       intervals in the future
        a_lower[(av > 1) & np.isnan(bv)] = one_plus_eps
        a_upper[(av > 1) & np.isnan(bv)] = np.inf
        a_lower[(av < 1) & np.isnan(bv)] = -np.inf
        a_upper[(av < 1) & np.isnan(bv)] = one_minus_eps

        # handle 0 ** bv
        # - +-0 ** +-0 = 1 -> discontinuous, force av = +-0 and bv = +-0
        # - +0 ** (>0) = +0 -> allow any bv > 0
        # - +0 ** (<0) = -inf -> allow any bv < 0
        # - -0 ** (>0) = +-0 -> too tricky, just keep both av and bv the same
        # - -0 ** (<0) = +-inf -> too tricky, just keep both av and bv the same
        np.copyto(a_lower, av, where=_is_negative_zero(av), casting="no")
        np.copyto(a_upper, av, where=_is_negative_zero(av), casting="no")
        np.copyto(b_lower, bv, where=_is_negative_zero(av), casting="no")
        np.copyto(b_upper, bv, where=_is_negative_zero(av), casting="no")
        a_lower[(av == 0) & (bv == 0)] = -0.0
        a_upper[(av == 0) & (bv == 0)] = +0.0
        b_lower[(av == 0) & (bv == 0)] = -0.0
        b_upper[(av == 0) & (bv == 0)] = +0.0
        b_lower[_is_positive_zero(av) & (bv > 0)] = smallest_subnormal
        b_upper[_is_positive_zero(av) & (bv > 0)] = np.inf
        b_lower[_is_positive_zero(av) & (bv < 0)] = -np.inf
        b_upper[_is_positive_zero(av) & (bv < 0)] = -smallest_subnormal

        # [-NaN, -inf, +inf, +NaN] ** +-0 = 1 -> so force bv = +-0 and allow
        #  any av
        a_lower[bv == 0] = -np.inf
        a_upper[bv == 0] = +np.inf
        b_lower[bv == 0] = -0.0
        b_upper[bv == 0] = +0.0

        # NaN ** (<0>) = NaN -> allow any av != 0
        # TODO: an interval union could represent that the two disjoint
        #       intervals in the future
        b_lower[np.isnan(av) & (bv > 0)] = smallest_subnormal
        b_upper[np.isnan(av) & (bv > 0)] = np.inf
        b_lower[np.isnan(av) & (bv < 0)] = -np.inf
        b_upper[np.isnan(av) & (bv < 0)] = -smallest_subnormal

        # powers of negative numbers are just too tricky, so force av and bv
        np.copyto(a_lower, av, where=(av < 0), casting="no")
        np.copyto(a_upper, av, where=(av < 0), casting="no")
        np.copyto(b_lower, bv, where=(av < 0), casting="no")
        np.copyto(b_upper, bv, where=(av < 0), casting="no")

        # flip a bounds if bv < 0; flip b bounds if av < 1
        # - av < 1 -> av ** b_lower >= av ** b_upper -> flip b bounds
        # - av < 1 & bv < 0 -> a_lower ** bv >= a_upper ** bv -> flip a bounds
        # - av < 1 & bv > 0 -> a_lower ** bv <= a_upper ** bv
        # - av > 1 -> av ** b_lower <= av ** b_upper
        # - av > 1 & bv < 0 -> a_lower ** bv >= a_upper ** bv -> flip a bounds
        # - av > 1 & bv > 0 -> a_lower ** bv <= a_upper ** bv
        # so that the below nudging works with the worst case combinations
        a_lower, a_upper = (
            _where(np.less(bv, 0), a_upper, a_lower),
            _where(np.less(bv, 0), a_lower, a_upper),
        )
        b_lower, b_upper = (
            _where(np.less(av, 1), b_upper, b_lower),
            _where(np.less(av, 1), b_lower, b_upper),
        )

        # stack the bounds on a and b so that we can nudge their bounds, if
        #  necessary, together
        tl_stack = _stack([a_lower, b_lower])
        tu_stack = _stack([a_upper, b_upper])

        def compute_term_power(
            t_stack: np.ndarray[tuple[int, Ps], np.dtype[F]],
        ) -> np.ndarray[tuple[int, Ps], np.dtype[F]]:
            total_power: np.ndarray[tuple[Ps], np.dtype[F]] = np.power(
                t_stack[0], t_stack[1]
            )

            return _broadcast_to(
                _ensure_array(total_power).reshape((1, *exprv.shape)),
                (t_stack.shape[0], *exprv.shape),
            )

        exprv = _ensure_array(exprv)
        expr_lower = _ensure_array(expr_lower)
        expr_upper = _ensure_array(expr_upper)

        tl_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_power,
            _broadcast_to(
                exprv.reshape((1, *exprv.shape)),
                (tl_stack.shape[0], *exprv.shape),
            ),
            _stack([av, bv]),
            tl_stack,
            _broadcast_to(
                expr_lower.reshape((1, *exprv.shape)),
                (tl_stack.shape[0], *exprv.shape),
            ),
            _broadcast_to(
                expr_upper.reshape((1, *exprv.shape)),
                (tl_stack.shape[0], *exprv.shape),
            ),
        )
        tu_stack = guarantee_stacked_arg_within_expr_bounds(
            compute_term_power,
            _broadcast_to(
                exprv.reshape((1, *exprv.shape)),
                (tu_stack.shape[0], *exprv.shape),
            ),
            _stack([av, bv]),
            tu_stack,
            _broadcast_to(
                expr_lower.reshape((1, *exprv.shape)),
                (tu_stack.shape[0], *exprv.shape),
            ),
            _broadcast_to(
                expr_upper.reshape((1, *exprv.shape)),
                (tu_stack.shape[0], *exprv.shape),
            ),
        )

        # extract the bounds for a and b and undo any earlier flips
        a_lower, a_upper = (
            _where(np.less(bv, 0), tu_stack[0], tl_stack[0]),
            _where(np.less(bv, 0), tl_stack[0], tu_stack[0]),
        )
        b_lower, b_upper = (
            _where(np.less(av, 1), tu_stack[1], tl_stack[1]),
            _where(np.less(av, 1), tl_stack[1], tu_stack[1]),
        )

        # recurse into a and b to propagate their bounds, then combine their
        #  bounds on Xs
        Xs_lower, Xs_upper = a.compute_data_bounds(
            a_lower,
            a_upper,
            Xs,
            late_bound,
        )

        bl, bu = b.compute_data_bounds(
            b_lower,
            b_upper,
            Xs,
            late_bound,
        )
        Xs_lower = _maximum_zero_sign_sensitive(Xs_lower, bl)
        Xs_upper = _minimum_zero_sign_sensitive(Xs_upper, bu)

        # ensure that the bounds on Xs include Xs
        Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
        Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} ** {self._b!r}"



import operator
from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _broadcast_to,
    _ensure_array,
    _is_negative_zero,
    _is_positive_zero,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _stack,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_stacked_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .abs import ScalarAbs
from .constfold import ScalarFoldedConstant
from .literal import Number
from .neg import ScalarNegate


class ScalarAdd(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    def __new__(cls, a: AnyExpr, b: AnyExpr) -> "ScalarAdd | Number":  # type: ignore[misc]
        ab = Number.symbolic_fold_binary(a, b, operator.add)
        if ab is not None:
            return ab
        return super().__new__(cls)

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarAdd | Number":
        return ScalarAdd(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a, self._b, dtype, np.add, ScalarAdd
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.add(self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        return compute_left_associate_sum_data_bounds(
            self, expr_lower, expr_upper, Xs, late_bound
        )

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} + {self._b!r}"


class ScalarSubtract(Expr[AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_a", "_b")
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, a: AnyExpr, b: AnyExpr) -> None:
        self._a = a
        self._b = b

    def __new__(cls, a: AnyExpr, b: AnyExpr) -> "ScalarSubtract | Number":  # type: ignore[misc]
        ab = Number.symbolic_fold_binary(a, b, operator.sub)
        if ab is not None:
            return ab
        return super().__new__(cls)

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr]:
        return (self._a, self._b)

    @override
    def with_args(self, a: AnyExpr, b: AnyExpr) -> "ScalarSubtract | Number":
        return ScalarSubtract(a, b)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_binary(
            self._a, self._b, dtype, np.subtract, ScalarSubtract
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.subtract(self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        return compute_left_associate_sum_data_bounds(
            self, expr_lower, expr_upper, Xs, late_bound
        )

    @override
    def __repr__(self) -> str:
        return f"{self._a!r} - {self._b!r}"


# left-associative sum over at least three terms, i.e. a + b + c = (a + b) + c
#  - zero terms: sum identity = 0
#  - one term: term
#  - two terms: binary sum
# this class avoids the deep nesting that's required to represent a
#  left-associative sum with ScalarAdd's
class ScalarLeftAssociativeSum(Expr[AnyExpr, AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]):
    __slots__: tuple[str, ...] = ("_a", "_b", "_c", "_ds")
    _a: AnyExpr
    _b: AnyExpr
    _c: AnyExpr
    _ds: tuple[AnyExpr, ...]

    def __new__(cls, *ts: AnyExpr) -> AnyExpr | Number:  # type: ignore[misc]
        # base case: sum identity
        if len(ts) == 0:
            return Number.ZERO

        # we can only fold consecutive symbolic integers from the left
        a, *tsr = ts
        i = 0
        broken = False
        for i, b in enumerate(tsr):
            ab = Number.symbolic_fold_binary(a, b, operator.add)
            if ab is None:
                broken = True
                break
            a = ab
        i += not broken

        match tsr[i:]:
            # addition over one term
            case ():
                return a
            # binary addition
            case (b,):
                return ScalarAdd(a, b)
            # sum over at least three terms
            case tsri:
                sum_ = super().__new__(cls)
                sum_._a = a
                sum_._b, sum_._c, *ds = tsri
                sum_._ds = tuple(ds)
                return sum_

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr, AnyExpr, *tuple[AnyExpr, ...]]:
        return (self._a, self._b, self._c, *self._ds)

    @override
    def with_args(  # type: ignore[override]
        self, a: AnyExpr, b: AnyExpr, c: AnyExpr, *ds: AnyExpr
    ) -> AnyExpr | Number:
        return ScalarLeftAssociativeSum(a, b, c, *ds)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        # we first individually fold each term
        facc = self._a.constant_fold(dtype)
        fts = [self._b.constant_fold(dtype), self._c.constant_fold(dtype)] + [
            d.constant_fold(dtype) for d in self._ds
        ]

        # next, we can only combine consecutive folded terms from the left
        i = 0
        broken = False
        for i, ft in enumerate(fts):
            if isinstance(facc, Expr) or isinstance(ft, Expr):
                broken = True
                break
            facc = np.add(facc, ft)
        i += not broken

        # if all were folded, return the folded constant
        if not broken:
            return facc

        # otherwise turn all folded constant terms back into expressions
        faccexpr = ScalarFoldedConstant.from_folded(facc)
        ftexprs = [ScalarFoldedConstant.from_folded(ft) for ft in fts[i:]]

        # and create a sum of the folded terms
        return ScalarLeftAssociativeSum(faccexpr, *ftexprs)

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        # evaluate the sum left-associative, i.e. a + b + c = (a + b) + c
        acc: np.ndarray[tuple[Ps], np.dtype[F]] = np.add(
            self._a.eval(Xs, late_bound), self._b.eval(Xs, late_bound)
        )

        acc += self._c.eval(Xs, late_bound)

        for d in self._ds:
            acc += d.eval(Xs, late_bound)

        return acc

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        return compute_left_associate_sum_data_bounds(
            self, expr_lower, expr_upper, Xs, late_bound
        )

    @override
    def __repr__(self) -> str:
        abc = f"{self._a!r} + {self._b!r} + {self._c!r}"

        if len(self._ds) <= 0:
            return abc

        return f"{abc} + {' + '.join(repr(d) for d in self._ds)}"


def compute_left_associate_sum_data_bounds(
    expr: ScalarAdd | ScalarSubtract | ScalarLeftAssociativeSum,
    expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
    expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
    Xs: np_sndarray[Ps, Ns, np.dtype[F]],
    late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
) -> tuple[
    np_sndarray[Ps, Ns, np.dtype[F]],
    np_sndarray[Ps, Ns, np.dtype[F]],
]:
    def _zero_add(
        a: np.ndarray[tuple[Ps], np.dtype[F]], b: np.ndarray[tuple[Ps], np.dtype[F]]
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        sum_: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(np.add(a, b))
        np.copyto(sum_, a, where=(b == 0), casting="no")
        return sum_

    left_associative_sum = as_left_associative_sum(expr)

    termvs: list[np.ndarray[tuple[Ps], np.dtype[F]]] = []
    abs_factorvs: list[None | np.ndarray[tuple[Ps], np.dtype[F]]] = []

    for term in left_associative_sum:
        abs_factor = get_expr_left_associative_abs_factor_approximate(term)
        termvs.append(term.eval(Xs, late_bound))
        abs_factorvs.append(
            None
            if abs_factor is None
            else np.multiply(
                abs_factor.eval(Xs, late_bound),
                _ensure_array(term.eval_has_data(Xs, late_bound)).astype(Xs.dtype),
            )
        )

    # evaluate the total expression sum
    exprv: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
        sum(termvs[1:], start=termvs[0])
    )
    expr_lower = _ensure_array(expr_lower)
    expr_upper = _ensure_array(expr_upper)

    # compute the sum of absolute factors
    total_abs_factor_: None | np.ndarray[tuple[Ps], np.dtype[F]] = None
    for abs_factorv in abs_factorvs:
        if abs_factorv is None:
            continue
        if total_abs_factor_ is None:
            total_abs_factor_ = _ensure_array(abs_factorv, copy=True)
        else:
            total_abs_factor_ += abs_factorv
    assert total_abs_factor_ is not None
    total_abs_factor: np.ndarray[tuple[Ps], np.dtype[F]] = total_abs_factor_

    # drop into expression difference bounds to divide up the bound
    # for NaN sums, we use a zero difference to ensure NaNs don't
    #  accidentally propagate into the term difference bounds
    expr_lower_diff: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
        np.subtract(expr_lower, exprv)
    )
    expr_lower_diff[np.isnan(expr_lower_diff)] = 0
    expr_upper_diff: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
        np.subtract(expr_upper, exprv)
    )
    expr_upper_diff[np.isnan(expr_upper_diff)] = 0

    # if exprv is NaN, expr_[lower|upper]_diff are zero
    # if expr_[lower|upper] is infinite but exprv is finite,
    #  expr_[lower|upper]_diff is the same infinity as expr_[lower|upper]
    # if expr_[lower|upper] is infinite and exprv is infinite,
    #  (a) -inf - -inf and inf - inf are NaN, so expr_[lower|upper]_diff is
    #      zero
    #  (b) -inf - inf and inf - -inf are +-inf, so expr_[lower|upper]_diff is
    #      the same infinity as expr_[lower|upper]
    tfl: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
        np.divide(expr_lower_diff, total_abs_factor)
    )
    tfu: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
        np.divide(expr_upper_diff, total_abs_factor)
    )

    fmax = np.finfo(Xs.dtype).max

    # ensure that the bounds never contain both -inf and +inf since that would
    #  allow NaN to sneak in
    inf_clash = (tfl == -np.inf) & (tfu == np.inf)
    tfl[inf_clash] = -fmax
    tfu[inf_clash] = fmax

    any_nan: np.ndarray[tuple[Ps], np.dtype[np.bool]] = np.isnan(termvs[0])
    for termv in termvs[1:]:
        any_nan |= np.isnan(termv)

    # stack the lower and upper bounds for each term factor
    # if total_abs_factor or exprv is non-finite:
    #  - non-finite values must be preserved exactly
    #  - if there is any NaN term, finite values can have any finite value
    #  - otherwise finite values are also preserved exactly, for simplicity
    # if total_abs_factor is zero, all abs_factorv are also zero and the term
    #  should keep the same value (unless the term is finite and there is any
    #  NaN term, see above)
    # otherwise we split up the difference bound for the terms by their factors
    tl_stack_: list[np.ndarray[tuple[Ps], np.dtype[F]]] = []
    for termv, abs_factorv in zip(termvs, abs_factorvs):
        if abs_factorv is None:
            continue
        tl: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(termv, copy=True)
        tl[np.isfinite(termv) & any_nan] = -fmax
        np.copyto(
            tl,
            _zero_add(termv, tfl * abs_factorv),
            where=(
                (total_abs_factor != 0)
                & np.isfinite(total_abs_factor)
                & np.isfinite(exprv)
            ),
            casting="no",
        )
        # ensure that finite terms don't overflow into infinity unless allowed
        #  by the difference bounds
        tl[
            (tl == -np.inf)
            & np.isfinite(termv)
            & np.isfinite(tfl)
            & np.isfinite(abs_factorv)
        ] = -fmax
        # optimistically try to include -0.0
        tl[(tl == 0) & ~_is_positive_zero(expr_lower)] = -0.0
        tl_stack_.append(tl)
    tl_stack = _stack(tl_stack_)
    tu_stack_: list[np.ndarray[tuple[Ps], np.dtype[F]]] = []
    for termv, abs_factorv in zip(termvs, abs_factorvs):
        if abs_factorv is None:
            continue
        tu: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(termv, copy=True)
        tu[np.isfinite(termv) & any_nan] = fmax
        np.copyto(
            tu,
            _zero_add(termv, tfu * abs_factorv),
            where=(
                (total_abs_factor != 0)
                & np.isfinite(total_abs_factor)
                & np.isfinite(exprv)
            ),
            casting="no",
        )
        # ensure that finite terms don't overflow into infinity unless allowed
        #  by the difference bounds
        tu[
            (tu == np.inf)
            & np.isfinite(termv)
            & np.isfinite(tfl)
            & np.isfinite(abs_factorv)
        ] = fmax
        # optimistically try to include +0.0
        tu[(tu == 0) & ~_is_negative_zero(expr_upper)] = +0.0
        tu_stack_.append(tu)
    tu_stack = _stack(tu_stack_)

    def compute_term_sum(
        t_stack: np.ndarray[tuple[int, Ps], np.dtype[F]],
    ) -> np.ndarray[tuple[int, Ps], np.dtype[F]]:
        total_sum: None | np.ndarray[tuple[Ps], np.dtype[F]] = None
        i = 0

        for termv, abs_factorv in zip(termvs, abs_factorvs):
            if total_sum is None:
                if abs_factorv is None:
                    total_sum = _ensure_array(termv, copy=True)
                else:
                    total_sum = _ensure_array(t_stack[i], copy=True)
            elif abs_factorv is None:
                total_sum += termv
            else:
                total_sum += t_stack[i]
            i += abs_factorv is not None

        assert total_sum is not None

        return _broadcast_to(
            _ensure_array(total_sum).reshape((1, *exprv.shape)),
            (t_stack.shape[0], *exprv.shape),
        )

    # handle rounding errors in the total absolute factor early
    tl_stack = guarantee_stacked_arg_within_expr_bounds(
        compute_term_sum,
        _broadcast_to(
            exprv.reshape((1, *exprv.shape)), (tl_stack.shape[0], *exprv.shape)
        ),
        _stack(
            [
                termv
                for termv, abs_factorv in zip(termvs, abs_factorvs)
                if abs_factorv is not None
            ]
        ),
        tl_stack,
        _broadcast_to(
            expr_lower.reshape((1, *exprv.shape)),
            (tl_stack.shape[0], *exprv.shape),
        ),
        _broadcast_to(
            expr_upper.reshape((1, *exprv.shape)),
            (tl_stack.shape[0], *exprv.shape),
        ),
    )
    tu_stack = guarantee_stacked_arg_within_expr_bounds(
        compute_term_sum,
        _broadcast_to(
            exprv.reshape((1, *exprv.shape)), (tu_stack.shape[0], *exprv.shape)
        ),
        _stack(
            [
                termv
                for termv, abs_factorv in zip(termvs, abs_factorvs)
                if abs_factorv is not None
            ]
        ),
        tu_stack,
        _broadcast_to(
            expr_lower.reshape((1, *exprv.shape)),
            (tu_stack.shape[0], *exprv.shape),
        ),
        _broadcast_to(
            expr_upper.reshape((1, *exprv.shape)),
            (tu_stack.shape[0], *exprv.shape),
        ),
    )

    xl: np_sndarray[Ps, Ns, np.dtype[F]]
    xu: np_sndarray[Ps, Ns, np.dtype[F]]
    Xs_lower_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
    Xs_upper_: None | np_sndarray[Ps, Ns, np.dtype[F]] = None
    i = 0
    for term, abs_factorv in zip(left_associative_sum, abs_factorvs):
        if abs_factorv is None:
            continue

        # recurse into the terms with a weighted bound
        xl, xu = term.compute_data_bounds(
            tl_stack[i],
            tu_stack[i],
            Xs,
            late_bound,
        )

        # combine the inner data bounds
        if Xs_lower_ is None:
            Xs_lower_ = xl
        else:
            Xs_lower_ = _maximum_zero_sign_sensitive(Xs_lower_, xl)
        if Xs_upper_ is None:
            Xs_upper_ = xu
        else:
            Xs_upper_ = _minimum_zero_sign_sensitive(Xs_upper_, xu)

        i += 1

    assert Xs_lower_ is not None
    assert Xs_upper_ is not None
    Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_lower_
    Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = Xs_upper_

    Xs_lower = _minimum_zero_sign_sensitive(Xs_lower, Xs)
    Xs_upper = _maximum_zero_sign_sensitive(Xs_upper, Xs)

    return Xs_lower, Xs_upper


def as_left_associative_sum(
    expr: ScalarAdd | ScalarSubtract | ScalarLeftAssociativeSum,
) -> tuple[AnyExpr, ...]:
    terms_rev: list[AnyExpr] = []

    while True:
        # a left associative sum ((a + b) + c) + d
        # has the reverse stack d, c, b, a
        if isinstance(expr, ScalarLeftAssociativeSum):
            terms_rev.extend(expr._ds[::-1])
            terms_rev.append(expr._c)

        # rewrite ( a - b ) as ( a + (-b) ), which is bitwsie equivalent for
        #  floating-point numbers
        terms_rev.append(
            ScalarNegate(expr._b) if isinstance(expr, ScalarSubtract) else expr._b
        )

        if isinstance(expr._a, ScalarAdd | ScalarSubtract | ScalarLeftAssociativeSum):
            expr = expr._a
        else:
            terms_rev.append(expr._a)
            break

    return tuple(terms_rev[::-1])


def get_expr_left_associative_abs_factor_approximate(expr: AnyExpr) -> None | AnyExpr:
    from .divmul import ScalarDivide, ScalarMultiply  # noqa: PLC0415

    if not expr.has_data:
        return None

    if not isinstance(expr, ScalarMultiply | ScalarDivide):
        return Number.ONE

    factor_stack: list[tuple[AnyExpr, type[ScalarMultiply] | type[ScalarDivide]]] = []

    while True:
        factor_stack.append((expr._b, type(expr)))

        if isinstance(expr._a, ScalarMultiply | ScalarDivide):
            expr = expr._a
        else:
            factor_stack.append(
                (Number.ONE if expr._a.has_data else expr._a, ScalarMultiply)
            )
            break

    while len(factor_stack) > 1:
        (a, _), (b, ty) = factor_stack.pop(), factor_stack.pop()
        factor_stack.append(
            (
                ty(a, Number.ONE if b.has_data else b),
                ScalarMultiply,
            )
        )

    [(factor, _)] = factor_stack

    return ScalarAbs(factor)



"""
Abstract base class for the safeguards.
"""

__all__ = ["Safeguard"]

from abc import ABC, abstractmethod
from collections.abc import Set
from typing import ClassVar, Literal, Self

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ..utils.bindings import Bindings, Parameter
from ..utils.typing import JSON, S, T


class Safeguard(ABC):
    """
    Safeguard abstract base class.
    """

    __slots__: tuple[str, ...] = ()
    kind: ClassVar[str]
    """Safeguard kind."""

    @property
    @abstractmethod
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return frozenset()

    @abstractmethod
    def check(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if the `prediction` array upholds the property enforced by this
        safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.
        """
        pass

    @abstractmethod
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        The config must include a 'kind' field with the safeguard kind. All
        values must be compatible with JSON encoding.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        pass

    @classmethod
    def from_config(cls, config: dict[str, JSON]) -> Self:
        """
        Instantiate the safeguard from a configuration [`dict`][dict].

        Parameters
        ----------
        config : dict[str, JSON]
            Configuration of the safeguard.

        Returns
        -------
        safeguard : Self
            Instantiated safeguard.
        """

        return cls(**config)

    @override
    def __repr__(self) -> str:
        config: dict[str, JSON] = {
            k: v for k, v in self.get_config().items() if k != "kind"
        }

        return f"{type(self).__name__}({', '.join(f'{k}={v!r}' for k, v in config.items())})"



"""
Abstract base class for the stencil safeguards.
"""

__all__ = ["StencilSafeguard"]

from abc import ABC, abstractmethod
from typing import Literal, final

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils.bindings import Bindings
from ...utils.intervals import IntervalUnion
from ...utils.typing import S, T
from ..abc import Safeguard
from . import BoundaryCondition, NeighbourhoodAxis


class StencilSafeguard(Safeguard, ABC):
    """
    Stencil safeguard abstract base class.

    Stencil safeguards describe properties that are computed in a
    neighbourhood around each element, i.e. whether or not an element satisfies
    the safeguard is coupled to its neighbouring elements.
    """

    __slots__: tuple[str, ...] = ()

    @abstractmethod
    def compute_check_neighbourhood_for_data_shape(
        self, data_shape: tuple[int, ...]
    ) -> tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]:
        """
        Compute the shape of the data neighbourhood for data of a given shape.
        Boundary conditions of the same kind are combined, but separate kinds
        are tracked separately.

        An empty [`dict`][dict] is returned along dimensions for which the
        stencil safeguard does not need to look at adjacent data points.

        This method also checks that the data shape is compatible with this
        stencil safeguard.

        Parameters
        ----------
        data_shape : tuple[int, ...]
            The shape of the data.

        Returns
        -------
        neighbourhood_shape : tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]
            The shape of the data neighbourhood.
        """

        pass

    @final
    @override
    def check(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if the `prediction` array upholds the property enforced by this
        safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.
        """

        return bool(
            np.all(
                self.check_pointwise(
                    data, prediction, late_bound=late_bound, where=where
                )
            )
        )

    @abstractmethod
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements in the `prediction` array uphold the neighbourhood
        property enforced by this safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.
        """

        pass

    @abstractmethod
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the safeguard's guarantees with respect
        to the `data` are upheld.

        The returned union of intervals must not have any overlap between the
        intervals inside the union. The `data` must be contained in the union.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the safeguard's guarantees are upheld.
        """

        pass

    @abstractmethod
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        For stencil safeguards, the footprint usually extends beyond
        `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

            Conceptually, `where` is applied to `footprint` at the end.

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        pass

    @abstractmethod
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        For stencil safeguards, the inverse footprint usually extends beyond
        `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

            Conceptually, `where` is applied to `foot` at the start.

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        pass



"""
Utility functions to cast arrays to floating-point, binary, and total-order representations.
"""

__all__ = [
    "ToFloatMode",
    "to_float",
    "from_float",
    "as_bits",
    "to_total_order",
    "from_total_order",
    "lossless_cast",
    "saturating_finite_float_cast",
]

from enum import Enum, auto
from typing import assert_never

import numpy as np

from ._compat import _ensure_array, _is_of_dtype
from ._float128 import _float128_dtype
from .error import TypeSetError, ctx
from .typing import F, S, T, U


class ToFloatMode(Enum):
    """
    Mode for losslessly converting numeric data to floating-point.
    """

    lossless = auto()
    """
    Automatically select the smallest floating-point data type that can
    losslessly represent the input data.
    """

    float16 = auto()
    """
    Use 16 bit floating-point precision, if lossless.
    """

    float32 = auto()
    """
    Use 32 bit floating-point precision, if lossless.
    """

    float64 = auto()
    """
    Use 64 bit floating-point precision, if lossless.
    """

    float128 = auto()
    """
    Use 128 bit floating-point precision, if lossless.
    """

    def floating_point_dtype_for(self, dtype: np.dtype[T]) -> np.dtype[np.floating]:
        """
        Select the floating-point dtype for the input `dtype`.

        Only data type supported by the safeguards (see
        [`Safeguards.supported_dtypes`][.....api.Safeguards.supported_dtypes])
        are supported by this method.

        The [`numpy_quaddtype`](https://pypi.org/project/numpy-quaddtype/)
        package is used to provide a true 128 bit floating-point data type.

        Parameters
        ----------
        dtype : np.dtype[T]
            Data type of the input data.

        Returns
        -------
        ftype : np.dtype[np.floating]
            Floating-point data type that can losslessly represent all values
            from the input `dtype`.

        Raises
        ------
        TypeError
            if `dtype` cannot be losslessly cast to `self`.
        """

        match self:
            case ToFloatMode.lossless:
                SMALLEST_LOSSLESS_FTYPE: dict[
                    np.dtype[np.number], np.dtype[np.floating]
                ] = {
                    np.dtype(np.int8): np.dtype(np.float16),
                    np.dtype(np.int16): np.dtype(np.float32),
                    np.dtype(np.int32): np.dtype(np.float64),
                    np.dtype(np.int64): _float128_dtype,
                    np.dtype(np.uint8): np.dtype(np.float16),
                    np.dtype(np.uint16): np.dtype(np.float32),
                    np.dtype(np.uint32): np.dtype(np.float64),
                    np.dtype(np.uint64): _float128_dtype,
                    np.dtype(np.float16): np.dtype(np.float16),
                    np.dtype(np.float32): np.dtype(np.float32),
                    np.dtype(np.float64): np.dtype(np.float64),
                }
                return SMALLEST_LOSSLESS_FTYPE[dtype]
            case ToFloatMode.float16:
                if dtype in (
                    np.dtype(np.int8),
                    np.dtype(np.uint8),
                    np.dtype(np.float16),
                ):
                    return np.dtype(np.float16)
                raise (
                    TypeError(f"cannot losslessly cast {dtype.name} to float16") | ctx
                )
            case ToFloatMode.float32:
                if dtype in (
                    np.dtype(np.int8),
                    np.dtype(np.uint8),
                    np.dtype(np.int16),
                    np.dtype(np.uint16),
                    np.dtype(np.float16),
                    np.dtype(np.float32),
                ):
                    return np.dtype(np.float32)
                raise (
                    TypeError(f"cannot losslessly cast {dtype.name} to float32") | ctx
                )
            case ToFloatMode.float64:
                if dtype in (
                    np.dtype(np.int8),
                    np.dtype(np.uint8),
                    np.dtype(np.int16),
                    np.dtype(np.uint16),
                    np.dtype(np.int32),
                    np.dtype(np.uint32),
                    np.dtype(np.float16),
                    np.dtype(np.float32),
                    np.dtype(np.float64),
                ):
                    return np.dtype(np.float64)
                raise (
                    TypeError(f"cannot losslessly cast {dtype.name} to float64") | ctx
                )
            case ToFloatMode.float128:
                if dtype in (
                    np.dtype(np.int8),
                    np.dtype(np.uint8),
                    np.dtype(np.int16),
                    np.dtype(np.uint16),
                    np.dtype(np.int32),
                    np.dtype(np.uint32),
                    np.dtype(np.int64),
                    np.dtype(np.uint64),
                    np.dtype(np.float16),
                    np.dtype(np.float32),
                    np.dtype(np.float64),
                ):
                    return _float128_dtype
                raise (
                    TypeError(f"cannot losslessly cast {dtype.name} to float128") | ctx
                )
            case _:
                assert_never(self)


def to_float(
    x: np.ndarray[S, np.dtype[T]], ftype: np.dtype[F]
) -> np.ndarray[S, np.dtype[F]]:
    """
    Losslessly convert the array `x` to the floating-point data type `ftype`.

    `ftype` must be a floating-point data type that can represent all values
    of the data type of `x` without loss in precision:

    * For floating-point data, it is at least the input data type.

    * For integer data, it is a floating-point type with sufficient precision
    to represent all integer values, i.e. a type whose mantissa has more bits
    than the integer type. For the supported floating-point types, this
    corresponds to choosing a floating-point data type with a larger bit width
    (e.g. at least [`np.float64`][numpy.float64] for [`np.int32`][numpy.int32]
    or [`np.uint32`][numpy.uint32] data).

    The
    [`ToFloatMode.floating_point_dtype_for`][..ToFloatMode.floating_point_dtype_for]
    method can be used to select a floating-point data type that fits the above
    criteria.

    Parameters
    ----------
    x : np.ndarray[S, np.dtype[T]]
        The array to convert.
    ftype : np.dtype[F]
        The floating-point data type to convert to, which must be able to
        losslessly represent all values of the data type of `x`.

    Returns
    -------
    converted : np.ndarray[S, np.dtype[F]]
        The converted array with the chosen floating-point data type.
    """

    assert np.issubdtype(ftype, np.floating)

    if np.issubdtype(x.dtype, np.floating):
        assert ftype.itemsize >= x.dtype.itemsize
    else:
        assert np.finfo(ftype).nmant >= x.dtype.itemsize

    if _is_of_dtype(x, ftype):
        return x

    # FIXME: https://github.com/numpy/numpy-user-dtypes/issues/163
    with np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore"):
        # lossless cast to floating-point data type with a sufficiently large
        #  mantissa
        return x.astype(ftype, casting="safe")


def from_float(
    x: np.ndarray[S, np.dtype[F]], dtype: np.dtype[T]
) -> np.ndarray[S, np.dtype[T]]:
    """
    Reverses the conversion of the array `x`, converted using the
    [`to_float`][..to_float], back to the original `dtype`.

    If the original `dtype` was floating-point with lower precision, the
    conversion is lossy.

    If the original `dtype` was integer, the rounding conversion is lossy.
    Infinite values are clamped to the minimum/maximum integer values.

    Parameters
    ----------
    x : np.ndarray[S, np.dtype[F]]
        The floating-point array to re-convert.
    dtype : np.dtype[T]
        The original dtype.

    Returns
    -------
    converted : np.ndarray[S, np.dtype[T]]
        The re-converted array with the original `dtype`.
    """

    x = _ensure_array(x)

    assert np.issubdtype(x.dtype, np.floating)

    if _is_of_dtype(x, dtype):
        return x

    if np.issubdtype(dtype, np.floating):
        with np.errstate(
            divide="ignore", over="ignore", under="ignore", invalid="ignore"
        ):
            # lossy cast to lower-precision floating-point number
            return x.astype(dtype, casting="unsafe")

    info = np.iinfo(dtype)  # type: ignore
    imin, imax = np.array(info.min, dtype=dtype), np.array(info.max, dtype=dtype)

    with np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore"):
        # lossy cast from floating-point to integer
        # round first with rint (round to nearest, ties to nearest even)
        converted: np.ndarray[S, np.dtype[T]] = _ensure_array(np.rint(x)).astype(
            dtype, casting="unsafe"
        )
        converted[np.greater(x, imax.astype(x.dtype, casting="safe"))] = imax
        converted[np.less(x, imin.astype(x.dtype, casting="safe"))] = imin

    return converted


def as_bits(
    a: np.ndarray[S, np.dtype[T]],
) -> np.ndarray[S, np.dtype[U]]:
    """
    Reinterprets the array `a` to its binary unsigned integer representation.

    Parameters
    ----------
    a : np.ndarray[S, np.dtype[T]]
        The array to reinterpret as binary.

    Returns
    -------
    binary : np.ndarray[S, np.dtype[U]]
        The binary unsigned integer representation of the array `a`.
    """

    return a.view(a.dtype.str.replace("f", "u").replace("i", "u"))


def to_total_order(a: np.ndarray[S, np.dtype[T]]) -> np.ndarray[S, np.dtype[U]]:
    """
    Reinterprets the array `a` to its total-order unsigned binary
    representation.

    In their total-order representation, the smallest value is mapped to
    unsigned zero, and the largest value is mapped to the largest unsigned
    value.

    For floating-point values, this implementation is based on Michael Herf's
    `FloatFlip` function, see <http://stereopsis.com/radix.html>.

    Parameters
    ----------
    a : np.ndarray[S, np.dtype[T]]
        The array to reinterpret to its total-order.

    Returns
    -------
    ordered : np.ndarray[S, np.dtype[U]]
        The total-order unsigned binary representation of the array `a`.

    Raises
    ------
    TypeSetError
        if `a` is not a signed/unsigned integer or floating array.
    """

    TypeSetError.check_or_raise(
        a.dtype.type, np.unsignedinteger | np.signedinteger | np.floating
    )

    if np.issubdtype(a.dtype, np.unsignedinteger):
        return a  # type: ignore

    utype = np.dtype(a.dtype.str.replace("i", "u").replace("f", "u"))

    if np.issubdtype(a.dtype, np.signedinteger):
        shift = np.iinfo(a.dtype).max  # type: ignore
        with np.errstate(
            over="ignore",
            under="ignore",
        ):
            return a.view(utype) + utype.type(shift) + utype.type(1)

    itype = a.dtype.str.replace("f", "i")
    bits = np.iinfo(utype).bits

    mask = (-((a.view(dtype=utype) >> (bits - 1)).view(dtype=itype))).view(
        dtype=utype
    ) | (utype.type(1) << (bits - 1))

    return a.view(dtype=utype) ^ mask


def from_total_order(
    a: np.ndarray[S, np.dtype[U]], dtype: np.dtype[T]
) -> np.ndarray[S, np.dtype[T]]:
    """
    Reverses the reinterpretation of the array `a` back from total-order
    unsigned binary to the provided `dtype`.

    For floating-point values, this implementation is based on Michael Herf's
    `IFloatFlip` function, see <http://stereopsis.com/radix.html>.

    Parameters
    ----------
    a : np.ndarray[S, np.dtype[U]]
        The array to reverse-reinterpret back from its total-order.

    Returns
    -------
    array : np.ndarray[S, np.dtype[T]]
        The array with its original `dtype`.

    Raises
    ------
    TypeSetError
        if `a` is not an unsigned integer array.
    TypeSetError
        if `dtype` is not a signed/unsigned integer or floating data type.
    """

    TypeSetError.check_or_raise(a.dtype.type, np.unsignedinteger)
    TypeSetError.check_or_raise(
        dtype.type, np.unsignedinteger | np.signedinteger | np.floating
    )

    if np.issubdtype(dtype, np.unsignedinteger):
        return a  # type: ignore

    if np.issubdtype(dtype, np.signedinteger):
        shift = np.array(np.iinfo(dtype).max, dtype=dtype)  # type: ignore
        with np.errstate(
            over="ignore",
            under="ignore",
        ):
            return a.view(dtype) + shift + dtype.type(1)

    utype = np.dtype(dtype.str.replace("f", "u"))
    itype = dtype.str.replace("f", "i")
    bits = np.iinfo(utype).bits

    mask = ((a >> (bits - 1)).view(dtype=itype) - 1).view(dtype=utype) | (
        utype.type(1) << (bits - 1)
    )

    return (a ^ mask).view(dtype=dtype)


def lossless_cast(
    x: int | float | np.number | np.ndarray[S, np.dtype[np.number]],
    dtype: np.dtype[T],
) -> np.ndarray[tuple[()] | S, np.dtype[T]]:
    """
    Try to losslessly convert `x` to the provided `dtype`.

    A lossless conversion is one that can be reversed while preserving the
    original value. Integer values can be losslessly converted to integer or
    floating-point types with sufficient precision. Floating-point values
    can only be converted to floating-point types.

    Parameters
    ----------
    x : int | float | np.ndarray[S, np.dtype[np.number]]
        The value or array to convert.
    dtype : np.dtype[T]
        The dtype to which the value or array should be converted.

    Returns
    -------
    converted : np.narray[tuple[()] | S, np.dtype[T]]
        The losslessly converted value or array with the given `dtype`.

    Raises
    ------
    TypeError
        if floating-point values are converted to an integer `dtype`.
    ValueError
        if not all values could be losslessly converted to `dtype`.
    """

    xa = np.array(x, copy=None)
    dtype_from = xa.dtype

    if np.issubdtype(dtype_from, np.floating) and not np.issubdtype(dtype, np.floating):
        raise TypeError(f"cannot losslessly cast from {dtype_from} to {dtype}") | ctx

    # we use unsafe casts here since we later check them for safety
    with np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore"):
        xa_to = _ensure_array(xa).astype(dtype, casting="unsafe")
        xa_back = xa_to.astype(dtype_from, casting="unsafe")

    lossless_same = xa == xa_back
    lossless_same |= np.isnan(xa) & np.isnan(xa_back)

    if isinstance(x, int):
        lossless_same &= (x < 0) == np.signbit(xa_back)
    else:
        lossless_same &= np.signbit(xa) == np.signbit(xa_back)

    if not np.all(lossless_same):
        raise (
            ValueError(
                f"cannot losslessly cast (some) values from {dtype_from} to "
                + f"{dtype}"
            )
            | ctx
        )

    return xa_to


def saturating_finite_float_cast(
    x: int | float | np.number | np.ndarray[S, np.dtype[np.number]],
    dtype: np.dtype[F],
) -> np.ndarray[tuple[()] | S, np.dtype[F]]:
    """
    Try to convert the finite `x` to the provided floating-point `dtype`.
    Under- and overflows are clamped to finite values.

    Parameters
    ----------
    x : int | float | np.ndarray[S, np.dtype[np.number]]
        The value or array to convert.
    dtype : np.dtype[F]
        The floating-point dtype to which the value or array should be
        converted.

    Returns
    -------
    converted : np.narray[tuple[()] | S, np.dtype[F]]
        The losslessly converted value or array with the given `dtype`.

    Raises
    ------
    ValueError
        if any values are non-finite, i.e. infinite or NaN.
    """

    assert np.issubdtype(dtype, np.floating)

    xa = np.array(x, copy=None)

    if not isinstance(x, int) and not np.all(np.isfinite(xa)):
        raise (
            ValueError(
                f"cannot cast non-finite values from {xa.dtype.name} to "
                + f"saturating finite {dtype.name}"
            )
            | ctx
        )

    # we use unsafe casts here since but are safe since
    # - we know that inputs are all finite
    # - we cast to float, where under- and overflows saturate to np.inf
    # - we later clamp the values to finite
    with np.errstate(divide="ignore", over="ignore", under="ignore", invalid="ignore"):
        xa_to = _ensure_array(xa).astype(dtype, casting="unsafe")

    # the above checks guarantee that there are no NaNs in xa
    if isinstance(x, int):
        assert (x < 0) == np.signbit(xa_to)
    else:
        assert np.all(np.signbit(xa) == np.signbit(xa_to))

    return np.nan_to_num(xa_to)



__all__ = [
    "T",
    "F",
    "Fi",
    "J",
    "Ps",
    "Ps2",
    "Ns",
    "np_sndarray",
    "np_sndarray2",
    "Ci",
    "Es",
    "coerce_to_flat",
]

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeAlias, TypeVar, TypeVarTuple

import numpy as np

T = TypeVar("T", bound=np.dtype[np.generic], covariant=True)
""" Any numpy [`dtype`][numpy.dtype] (covariant). """

F = TypeVar("F", bound=np.floating, covariant=True)
""" Any numpy [`floating`][numpy.floating]-point data type (covariant). """

Fi = TypeVar("Fi", bound=np.floating)
""" Any numpy [`floating`][numpy.floating]-point data type (invariant). """

J = TypeVar("J", bound=int, covariant=True)
""" Any [`int`][int] (covariant). """

Ps = TypeVar("Ps", bound=int, covariant=True)
""" Any flattened pointwise array shape [X.size] (covariant). """

Ps2 = TypeVar("Ps2", bound=tuple[int, ...], covariant=True)
""" Any pointwise array shape X.shape (covariant). """

Ns = TypeVar("Ns", bound=tuple[int, ...], covariant=True)
""" Any stencil neighbourhood array shape [*S.shape] (covariant). """

if TYPE_CHECKING:
    np_sndarray: TypeAlias = np.ndarray[tuple[Ps, *Ns], T]  # type: ignore
    """ Any stencil-extended [`np.ndarray[tuple[Ps, *Ns], T]`][numpy.ndarray]. """
else:
    # *TypeVar(bound=tuple) is not yet supported
    np_sndarray: TypeAlias = np.ndarray[tuple[Ps, Ns], T]  # type: ignore

if TYPE_CHECKING:
    np_sndarray2: TypeAlias = np.ndarray[tuple[*Ps2, *Ns], T]  # type: ignore
    """ Any stencil-extended [`np.ndarray[tuple[*Ps2, *Ns], T]`][numpy.ndarray]. """
else:
    # *TypeVar(bound=tuple), *TypeVar(bound=tuple) is not supported
    np_sndarray2: TypeAlias = np.ndarray[tuple[Ps2, Ns], T]  # type: ignore

Ci = TypeVar("Ci", bound=Callable)
""" Any callable type (invariant). """

# FIXME: actually bound the types to be Expr
# https://discuss.python.org/t/how-to-use-typevartuple/67502
Es = TypeVarTuple("Es")
""" Tuple of [`Expr`][...expr.abc.Expr]s. """


def coerce_to_flat(a: np_sndarray[Ps, tuple[()], T]) -> np.ndarray[tuple[Ps], T]:
    """
    Coerce `np.ndarray[tuple[Ps, *tuple[()]], T]` to `np.ndarray[tuple[Ps], T]`.
    """

    return a  # type: ignore



"""
Logical selector (switch case) combinator safeguard.
"""

__all__ = ["SelectSafeguard"]

from abc import ABC
from collections.abc import Collection, Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils.bindings import Bindings, Parameter
from ...utils.error import TypeCheckError, ctx
from ...utils.intervals import Interval, IntervalUnion
from ...utils.typing import JSON, S, T
from ..abc import Safeguard
from ..pointwise.abc import PointwiseSafeguard
from ..stencil import BoundaryCondition, NeighbourhoodAxis
from ..stencil.abc import StencilSafeguard


class SelectSafeguard(Safeguard):
    """
    The `SelectSafeguard` guarantees that, for each element, the guarantees of
    the pointwise selected safeguard are upheld. This combinator allows
    selecting between several safeguards with per-element granularity.

    This combinator can be used to describe simple regions of interest where
    different safeguards, e.g. with different error bounds, are applied to
    different parts of the data.

    At the moment, only pointwise and stencil safeguards and combinations
    thereof can be combined by this select-combinator. The combinator is a
    pointwise or a stencil safeguard, depending on the safeguards it combines.

    Parameters
    ----------
    selector : int | str | Parameter
        Literal index or late-bound parameter name that is used to select
        between the `safeguards`.
    safeguards : Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard]
        At least one safeguard configuration [`dict`][dict]s or already
        initialized [`PointwiseSafeguard`][....pointwise.abc.PointwiseSafeguard]
        or [`StencilSafeguard`][....stencil.abc.StencilSafeguard].

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    ValueError
        if the `safeguards` collection is empty.
    IndexError
        if the `selector` is a literal [`int`][int] index that is not within
        the `0 <= selector < len(safeguards)` bounds.
    ...
        if instantiating a safeguard raises an exception.
    """

    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "select"

    def __init__(
        self,
        *,
        selector: int | str | Parameter,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> None:
        pass

    def __new__(  # type: ignore
        cls,
        *,
        selector: int | str | Parameter,
        safeguards: Collection[dict[str, JSON] | PointwiseSafeguard | StencilSafeguard],
    ) -> "_SelectPointwiseSafeguard | _SelectStencilSafeguard":
        from ... import SafeguardKind  # noqa: PLC0415

        with ctx.safeguardty(cls):
            with ctx.parameter("selector"):
                TypeCheckError.check_instance_or_raise(selector, int | str | Parameter)
                selector = (
                    Parameter(selector) if isinstance(selector, str) else selector
                )

            with ctx.parameter("safeguards"):
                TypeCheckError.check_instance_or_raise(safeguards, Collection)

                if len(safeguards) <= 0:
                    raise (
                        ValueError("can only select over at least one safeguard") | ctx
                    )

                safeguards_: list[PointwiseSafeguard | StencilSafeguard] = []
                safeguard: dict[str, JSON] | Safeguard
                for i, safeguard in enumerate(safeguards):
                    with ctx.index(i):
                        TypeCheckError.check_instance_or_raise(
                            safeguard, dict | PointwiseSafeguard | StencilSafeguard
                        )
                        if isinstance(safeguard, dict):
                            safeguard = SafeguardKind.from_config(safeguard)
                        TypeCheckError.check_instance_or_raise(
                            safeguard, PointwiseSafeguard | StencilSafeguard
                        )
                        safeguards_.append(safeguard)  # type: ignore

            with ctx.parameter("selector"):
                if isinstance(selector, int) and (
                    (selector < 0) or (selector >= len(safeguards_))
                ):
                    raise IndexError("invalid index") | ctx

        pointwise_safeguards_: list[PointwiseSafeguard] = [
            safeguard
            for safeguard in safeguards_
            if isinstance(safeguard, PointwiseSafeguard)
        ]

        if len(pointwise_safeguards_) == len(safeguards_):
            return _SelectPointwiseSafeguard(selector, *pointwise_safeguards_)
        else:
            return _SelectStencilSafeguard(selector, *safeguards_)

    @property
    def selector(self) -> int | Parameter:  # type: ignore
        """
        The selector index or late-bound selector parameter.
        """
        ...

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:  # type: ignore
        """
        The set of safeguards between which this combinator selects.
        """

        ...

    @property
    @override
    def late_bound(self) -> Set[Parameter]:  # type: ignore
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        ...

    @override
    def check(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> bool:
        """
        Check if, for all elements, the selected safeguard succeed the check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : bool
            `True` if the check succeeded.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `selector`'s late-bound parameter is not in `late_bound`.
        ValueError
            if the late-bound `selector` could not be broadcast to the `data`'s
            shape.
        TypeError
            if the late-bound `selector` is floating-point.
        ValueError
            if not all late-bound `selector` values could be losslessly
            converted to integer indices.
        IndexError
            if the late-bound `selector` indices are invalid for the
            selected-over `safeguards`.
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def check_pointwise(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check for which elements the selected safeguard succeed the check.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `selector`'s late-bound parameter is not in `late_bound`.
        ValueError
            if the late-bound `selector` could not be broadcast to the `data`'s
            shape.
        TypeError
            if the late-bound `selector` is floating-point.
        ValueError
            if not all late-bound `selector` values could be losslessly
            converted to integer indices.
        IndexError
            if the late-bound `selector` indices are invalid for the
            selected-over `safeguards`.
        ...
            if checking a safeguard raises an exception.
        """

        ...

    def compute_safe_intervals(  # type: ignore
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the safe intervals for the selected safeguard.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            The safe intervals.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `selector`'s late-bound parameter is not in `late_bound`.
        ValueError
            if the late-bound `selector` could not be broadcast to the `data`'s
            shape.
        TypeError
            if the late-bound `selector` is floating-point.
        ValueError
            if not all late-bound `selector` values could be losslessly
            converted to integer indices.
        IndexError
            if the late-bound `selector` indices are invalid for the
            selected-over `safeguards`.
        ...
            if computing the safe intervals for a safeguard raises an exception.
        """

        ...

    def compute_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        ...

    def compute_inverse_footprint(  # type: ignore
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        ...

    @override
    def get_config(self) -> dict[str, JSON]:  # type: ignore
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict
            Configuration of the safeguard.
        """

        ...


class _SelectSafeguardBase(ABC):
    __slots__: tuple[str, ...] = ()

    kind: ClassVar[str] = "select"

    @property
    def selector(self) -> int | Parameter:
        return self._selector  # type: ignore

    @property
    def safeguards(self) -> tuple[PointwiseSafeguard | StencilSafeguard, ...]:
        return self._safeguards  # type: ignore

    @property
    def late_bound(self) -> Set[Parameter]:
        parameters = frozenset(b for s in self.safeguards for b in s.late_bound)

        if isinstance(self.selector, Parameter):
            parameters = parameters.union([self.selector])

        return parameters

    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        # if a single safeguard is selected, just forward to it
        if isinstance(self.selector, int):
            return self.safeguards[self.selector].check_pointwise(
                data, prediction, late_bound=late_bound, where=where
            )

        with ctx.safeguardty(SelectSafeguard), ctx.parameter("selector"):
            selector = late_bound.resolve_ndarray_with_lossless_cast(
                self.selector, data.shape, np.dtype(np.intp)
            )

            with ctx.late_bound_parameter(self.selector):
                if np.any(selector < 0) or np.any(selector >= len(self.safeguards)):
                    raise IndexError("invalid indices") | ctx

        # only perform the checks where selected
        oks = [
            safeguard.check_pointwise(
                data, prediction, late_bound=late_bound, where=(selector == i) & where
            )
            for i, safeguard in enumerate(self.safeguards)
        ]

        # choose the check result from the selected safeguard
        with (
            ctx.safeguardty(SelectSafeguard),
            ctx.parameter("selector"),
            ctx.late_bound_parameter(self.selector),
        ):
            return np.choose(selector, oks)  # type: ignore

    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        # if a single safeguard is selected, just forward to it
        if isinstance(self.selector, int):
            return self.safeguards[self.selector].compute_safe_intervals(
                data, late_bound=late_bound, where=where
            )

        with ctx.safeguardty(SelectSafeguard), ctx.parameter("selector"):
            selector = late_bound.resolve_ndarray_with_lossless_cast(
                self.selector, data.shape, np.dtype(np.intp)
            )

            with ctx.late_bound_parameter(self.selector):
                if np.any(selector < 0) or np.any(selector >= len(self.safeguards)):
                    raise IndexError("invalid indices") | ctx

        valid: IntervalUnion[T, int, int] = Interval.full_like(data).into_union()

        # compute the safe intervals only where needed and then intersect
        # the `where` condition ensures that we only compute restrictions for
        #  the data points for which a safeguard was selected (and that other
        #  points do not impose such requirements)
        # but that the computed requirements can extend beyond the pointwise
        for i, safeguard in enumerate(self.safeguards):
            where_i = (selector == i) & where
            if np.any(where_i):
                valid = valid.intersect(
                    safeguard.compute_safe_intervals(
                        data, late_bound=late_bound, where=where_i
                    )
                )

        return valid

    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        # if a single safeguard is selected, just forward to it
        if isinstance(self.selector, int):
            return self.safeguards[self.selector].compute_footprint(
                foot, late_bound=late_bound, where=where
            )

        with ctx.safeguardty(SelectSafeguard), ctx.parameter("selector"):
            selector = late_bound.resolve_ndarray_with_lossless_cast(
                self.selector, foot.shape, np.dtype(np.intp)
            )

            with ctx.late_bound_parameter(self.selector):
                if np.any(selector < 0) or np.any(selector >= len(self.safeguards)):
                    raise IndexError("invalid indices") | ctx

        footprint = np.zeros_like(foot)

        # per point, only the selected safeguard contributes to the combined
        #  footprint
        # the union of the footprints is taken since the per-point footprint
        #  from a stencil safeguard may extend to neighbouring points
        for i, safeguard in enumerate(self.safeguards):
            where_i = (selector == i) & where
            if np.any(where_i):
                footprint |= safeguard.compute_footprint(
                    foot, late_bound=late_bound, where=where_i
                )

        return footprint

    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        # if a single safeguard is selected, just forward to it
        if isinstance(self.selector, int):
            return self.safeguards[self.selector].compute_inverse_footprint(
                foot, late_bound=late_bound, where=where
            )

        with ctx.safeguardty(SelectSafeguard), ctx.parameter("selector"):
            selector = late_bound.resolve_ndarray_with_lossless_cast(
                self.selector, foot.shape, np.dtype(np.intp)
            )

            with ctx.late_bound_parameter(self.selector):
                if np.any(selector < 0) or np.any(selector >= len(self.safeguards)):
                    raise IndexError("invalid indices") | ctx

        footprint = np.zeros_like(foot)

        # per point, only the selected safeguard contributes to the combined
        #  footprint
        # the union of the footprints is taken since the per-point footprint
        #  from a stencil safeguard may extend to neighbouring points
        for i, safeguard in enumerate(self.safeguards):
            where_i = (selector == i) & where
            if np.any(where_i):
                footprint |= safeguard.compute_inverse_footprint(
                    foot, late_bound=late_bound, where=where_i
                )

        return footprint

    def get_config(self) -> dict[str, JSON]:
        return dict(
            kind=type(self).kind,
            selector=str(self.selector)
            if isinstance(self.selector, Parameter)
            else self.selector,
            safeguards=[safeguard.get_config() for safeguard in self.safeguards],
        )

    @override
    def __repr__(self) -> str:
        return f"{SelectSafeguard.__name__}(selector={self.selector!r}, safeguards={list(self.safeguards)!r})"


class _SelectPointwiseSafeguard(_SelectSafeguardBase, PointwiseSafeguard):
    __slots__: tuple[str, ...] = ("_selector", "_safeguards")
    _selector: int | Parameter
    _safeguards: tuple[PointwiseSafeguard, ...]

    def __init__(
        self, selector: int | Parameter, *safeguards: PointwiseSafeguard
    ) -> None:
        self._selector = selector

        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard), (
                f"{safeguard!r} is not a pointwise safeguard"
            )
        self._safeguards = safeguards


class _SelectStencilSafeguard(_SelectSafeguardBase, StencilSafeguard):
    __slots__: tuple[str, ...] = ("_selector", "_safeguards")
    _selector: int | Parameter
    _safeguards: tuple[PointwiseSafeguard | StencilSafeguard, ...]

    def __init__(
        self,
        selector: int | Parameter,
        *safeguards: PointwiseSafeguard | StencilSafeguard,
    ) -> None:
        self._selector = selector

        for safeguard in safeguards:
            assert isinstance(safeguard, PointwiseSafeguard | StencilSafeguard), (
                f"{safeguard!r} is not a pointwise or stencil safeguard"
            )
        self._safeguards = safeguards

    @override
    def compute_check_neighbourhood_for_data_shape(
        self, data_shape: tuple[int, ...]
    ) -> tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]:
        neighbourhood: list[dict[BoundaryCondition, NeighbourhoodAxis]] = [
            dict() for _ in data_shape
        ]

        for safeguard in self._safeguards:
            if not isinstance(safeguard, StencilSafeguard):
                continue

            safeguard_neighbourhood = (
                safeguard.compute_check_neighbourhood_for_data_shape(data_shape)
            )

            for i, sn in enumerate(safeguard_neighbourhood):
                ni = neighbourhood[i]

                for b, s in sn.items():
                    if b in ni:
                        neighbourhood[i][b] = NeighbourhoodAxis(
                            max(ni[b].before, s.before), max(ni[b].after, s.after)
                        )
                    else:
                        neighbourhood[i][b] = s

        return tuple(neighbourhood)
